from contextlib import suppress

from .event_handler import EventHandler


class EventHandlerFactoryBase:
    def build(self, event_handler, *, keys=None, event_handler_type=EventHandler):
        builder = EventHandlerBuilder(event_handler, keys, event_handler_type)
        server_connector = self._build_server_connector()
        real_event_handler = builder.build(server_connector)
        event_handler.server_connector = server_connector
        with suppress(AttributeError):
            event_handler.start()
        real_event_handler.start()
        return real_event_handler

    def _build_server_connector(self):
        raise NotImplementedError()


class EventHandlerBuilder:
    def __init__(self, event_handler, supplied_keys, event_handler_type):
        self._analyzer = EventHandlerAnalyzer(event_handler, supplied_keys)
        self._event_handler_type = event_handler_type

    def build(self, server_connector):
        event_handler = self._event_handler_type(server_connector)
        server_connector.subscribe(*self._try_get_keys(event_handler))
        event_handler.handle_event = self._analyzer.handle_event
        with suppress(AttributeError):
            event_handler.cleanup = self._analyzer.cleanup
        return event_handler

    def _try_get_keys(self, event_handler):
        try:
            return self._analyzer.keys
        except ValueError:
            with suppress(AttributeError):
                return event_handler.keys
            raise


class EventHandlerAnalyzer:
    def __init__(self, event_handler, supplied_keys):
        self._event_handler = event_handler
        self._supplied_keys = supplied_keys

    @property
    def keys(self):
        if self._supplied_keys is None:
            try:
                return self._event_handler.keys
            except AttributeError:
                raise ValueError('No keys supplied!')
        return self._supplied_keys

    @property
    def handle_event(self):
        try:
            return self._event_handler.handle_event
        except AttributeError:
            if callable(self._event_handler):
                return self._event_handler
            raise ValueError('Object must implement handle_event or be a callable!')

    @property
    def cleanup(self):
        return self._event_handler.cleanup
